from .lib import *
from .short_functions import *


## Linecut
def make_linecut(image, cut_height='full'):
    '''Takes in an image and sums all the flux within CUT_HEIGHT from the x axis onto the x axis.
    Returns the dictinary with keys LEFT and RIGHT that correspond to the left and right halves of the image.
    If CUTHEIGHT is 'FULL': squish whole image onto the x axis.
    All units in pixels.'''

    if type(image) is dict:
        return {
            'left': make_linecut(image['left'], cut_height)['left'],
            'right': make_linecut(image['right'], cut_height)['right']
            }

    ydim, xdim = image.shape # pixels
    cx, cy = xdim//2, ydim//2 # pixels

    if cut_height == 'full':
        right_segment = image[:, cx: ]
        if xdim % 2 == 0:
            left_segment = image[:, cx-1: :-1]
        else:
            # Include central point corresponding to star
            left_segment = image[:, cx: :-1]
    else:
        cut_height = int(round(cut_height))
        if xdim % 2 == 0:
            assert ydim % 2 == 0
            left_segment = image[cy-cut_height:cy+cut_height, cx-1: :-1]
            right_segment = image[cy-cut_height:cy+cut_height, cx: ]
        else:
            assert ydim % 2 == 1
            # Include central point corresponding to star
            left_segment = image[cy-cut_height:cy+cut_height+1, cx: :-1]
            right_segment = image[cy-cut_height:cy+cut_height+1, cx: ]

    linecut = {
        'left': np.sum(left_segment, 0),
        'right': np.sum(right_segment, 0)
        }

    return linecut


def bin_linecut(linecut, r_bounds):
    '''Discretises an array LINECUT by averaging the values within each bin defined by R_BOUNDS. R_BOUNDS defines the index boundaries of each bin.'''

    if type(linecut) is dict:
        return {
            'left': bin_linecut(linecut['left'], r_bounds),
            'right': bin_linecut(linecut['right'], r_bounds)
            }

    r_bounds = np.array(r_bounds)
    linecut = np.array(linecut)
    coords = np.arange(0, len(linecut), 1) + 0.5

    assert r_bounds[-1] <= coords[-1] + 0.5

    nbins = len(r_bounds) - 1
    binned_linecut = np.zeros(nbins)

    for i in range(nbins):
        ks = np.where((coords >= r_bounds[i]) & (coords < r_bounds[i+1]))
        binned_linecut[i] = np.mean(linecut[ks])

    return binned_linecut


def cut_rings(rings, cut_height='full'):
    '''Apply MAKE_LINECUT to each element of the list RINGS'''
    return np.array([make_linecut(rings[i], cut_height) for i in range(len(rings))])


def bin_rings(rings, r_bounds, cut_height='full'):
    '''Apply MAKE_LINECUT then BIN_LINECUT to each element of the array RINGS.
    Returns a dictionary with keys LEFT and RIGHT. '''
    x = np.array([bin_linecut(make_linecut(rings[i], cut_height), r_bounds) for i in range(len(rings))])
    nrings = len(r_bounds) - 1
    abrl = handlelr(np.zeros)([nrings, nrings])
    for i in range(nrings):
        abrl['left'][i] = x[i]['left']
        abrl['right'][i] = x[i]['right']
    return abrl



from .lib import *
from .short_functions import *
from .linecut_functions import *


## Making rings
def make_ring(R, inner_radius, outer_radius, height=10, inclination=0, dim=200, kernel=None, n_points_per_pixel=200):
    '''Returns an image of a ring (equivalently called an annulus). 
    All units are in pixels. 
    Input:
        R: a list containing the radius of each particle in the Monte-Carlo simulation. If this is input as NONE, then the code draws R from a uniform distribution. 
        INNER_RADIUS: inner boundary of ring. 
        OUTER_RADIUS: outer boundary of ring. 
        HEIGHT: standard deviation of the Gaussian distribution of the ring's vertical height. 
        INCLINATION: inclination for viewing the ring relative to face-on. 0 is face-on. 90 is edge-on. 
        DIM: dimensions of the image. Output image is always square. 
        KERNEL: the kernel to concolve the image with. None means don't convolve with anything. 
        N_POINTS_PER_PIXEL: how many points to add to the image per pixel if the ring were to be viewed face-on. The more points you add, the higher the accuracy of the image, but the slower the code will run. THIS MUST BE CONSISTENT WITH R IF R IS GIVEN. 
    '''
    
    if R is None:
        n_points = calculate_n_points(n_points_per_pixel, [inner_radius, outer_radius])[0]
        R = np.random.uniform(inner_radius, outer_radius, n_points)
        # Make face-on brightness uniform
        B_adjust = R / ((inner_radius + outer_radius)/2) / n_points_per_pixel
    else:
        n_points = len(R)
        B_adjust = 1/n_points_per_pixel
    
    Theta = np.random.uniform(0, 360, n_points)/180*np.pi
    Z = np.random.normal(0, height, n_points)
    
    X = R * np.cos(Theta)
    Y = R * np.sin(Theta)
    B = brightness(R) * B_adjust
    
    if inclination:
        zip = np.array((X, Y, Z))
        zip2 = rotate_x(inclination/180*np.pi) @ zip
    
        X, Y, Z = zip2[0], zip2[1], zip2[2]

    # Bin into image
    image = np.zeros([dim, dim])
    
    if dim % 2 == 0:
        image_X = (np.round(X + dim//2 - 0.5)).astype(int)
        image_Y = (np.round(Y + dim//2 - 0.5)).astype(int)
    else:
        image_X = (np.round(X + dim//2)).astype(int)
        image_Y = (np.round(Y + dim//2)).astype(int)
    
    k = np.where((image_X >= 0) & (image_Y >= 0))
    image_X = image_X[k]
    image_Y = image_Y[k]
    for i in range(n_points):
        try:
            image[image_Y[i], image_X[i]] += B[i]
        except IndexError:
            pass
    
    # Convolve ring image
    if kernel is not None:
        image = convolve(image, kernel)
    
    return image


def calculate_n_points(n_points_per_pixel, r_bounds):
    '''Calculate number of points required for each ring defined by R_BOUNDS in order to achieve N_POINTS_PER_PIXEL if the image was viewed face-on. '''
    
    nrings = len(r_bounds) - 1
    ring_area = [np.pi*(r_bounds[i+1]**2 - r_bounds[i]**2) for i in range(nrings)]
    points_per_ring = np.array([int(ring_area[i] * n_points_per_pixel) for i in range(nrings)])
    
    return points_per_ring


def generate_R(n_points_per_pixel, r_bounds):
    '''The R distribution is a bit tricky to make up, so can achieve it by chopping up a set of points drawn from a triangular distribution. 
    This function generates r-coordinates for points in order to achieve N_POINTS_PER_PIXEL if the image was viewed face-on.'''
    
    points_per_ring = calculate_n_points(n_points_per_pixel, r_bounds)
    nrings = len(r_bounds) - 1
    
    # Try more points than total points to make sure each ring has enough points to match points_per_ring
    try_n_points = int(1.5 * n_points_per_pixel
        * np.pi * r_bounds[-1]**2)
    
    # Add more points if needed
    add_n_points = int(0.2 * try_n_points)
    
    # Draw from a triangular distribution
    Rs = np.random.triangular(0, r_bounds[-1], r_bounds[-1], try_n_points)
    R_per_ring = []
    
    for iring in range(nrings):
        # Don't try too many times
        while len(Rs) <= 10 * try_n_points:
            
            # Pick out points in ring
            R = Rs[(Rs >= r_bounds[iring]) & (Rs < r_bounds[iring+1])]
            
            # Take first points_per_ring if enough
            if len(R) >= points_per_ring[iring]:
                R_per_ring.append(R[ :points_per_ring[iring]])
                break
            
            # Otherwise add more points
            else:
                add_points = np.random.triangular(r_bounds[0], r_bounds[-1], r_bounds[-1], add_n_points)
                Rs = np.append(Rs, add_points)
                print('Added points')
                if len(Rs) > 10 * try_n_points:
                    print('Added too many points!')
    
    return R_per_ring


def make_all_rings(r_bounds, heights, inclination, dim, n_points_per_pixel=200, kernel=None, verbose=True):
    '''Makes all rings defined by R_BOUNDS.
    Input variables are defined in MAKE_RING.'''
    
    # Adjust parameters
    nrings = len(r_bounds) - 1
    heights = expand(heights, nrings)
    
    # Get R coordinates
    R_per_ring = generate_R(n_points_per_pixel, r_bounds)
        
    # Storage parameters
    all_ring_images = np.zeros([nrings, dim, dim])
    
    # Make rings
    for iring in range(nrings-1, -1, -1):
        if verbose:
            print(iring, sep='', end=' ')
        ring_image = make_ring(R_per_ring[iring], None, None, heights[iring], inclination, dim, kernel, n_points_per_pixel)
        all_ring_images[iring] = ring_image
    
    return all_ring_images



## Other rings functions
def generate_r_bounds(nrings, right_edge_pixel, left_edge_pixel=0):
    '''Makes boundaries for NRINGS number of annuli between LEFT_EDGE_PIXEL and RIGHT_EDGE_PIXEL.
    Satisfied the conditions that make sure each annulus isn't too wide or narrow.'''
    assert right_edge_pixel > left_edge_pixel
    range_pixels = right_edge_pixel - left_edge_pixel
    
    if nrings <= 20 and range_pixels/nrings >= 5:
        width_floor = 0.3 * range_pixels / nrings
        width_ceil = 2 * range_pixels / nrings
    else:
        width_floor = 0.3 * range_pixels / nrings
        width_ceil = 5 * range_pixels / nrings
    
    while True:
        r_bounds = np.random.uniform(left_edge_pixel, right_edge_pixel, nrings-1)
        r_bounds.sort()
        r_bounds = np.r_[left_edge_pixel, r_bounds, right_edge_pixel]
        r_widths = np.diff(r_bounds)
        if len(np.where((r_widths > width_ceil)|(r_widths < width_floor))[0]) == 0:
            break
    return r_bounds


def get_r_bounds(nrings, right_edge_pixel, n_iterations, timeit=True, use_flux_range=False):
    '''Reads in pre-generated R_BOUNDS if it exists. Otherwise makes a new set of R_BOUNDS with the right conditions and stores it. '''
    
    # Try stored rbounds
    import pickle
    prefix = 'Rbounds_'
    suffix = '.python'
    found_cache = 0
    
    if use_flux_range:
        assert len(right_edge_pixel) == 2
        if right_edge_pixel[0] == 0:
            left_edge_pixel = 0
            right_edge_pixel = right_edge_pixel[1]
            pixel_range = right_edge_pixel
        else:
            pixel_range = right_edge_pixel
            left_edge_pixel, right_edge_pixel = pixel_range
    else:
        left_edge_pixel = 0
        pixel_range = right_edge_pixel
    
    for filename in os.listdir():
        if prefix in filename:
            
            # Remove prefix and suffix and get parameters
            filename2 = filename.replace(prefix, '').replace(suffix, '')
            nrings2, pixel_range2, n_interations2 = filename2.split('_')
            
            # Compare parameters
            if nrings == float(nrings2) and str(pixel_range) == pixel_range2 and n_iterations <= float(n_interations2):
                print('    Found stored rbounds')
                found_cache = 1
                file = open(filename, 'rb')
                R_BOUNDS = pickle.load(file)
                file.close()
                break
    
    # Otherwise make rbounds
    if not found_cache:
        print('    Making rbounds')
        t0 = time.time()
        
        R_BOUNDS = np.zeros([n_iterations, nrings+1])
        for i in range(n_iterations):
            R_BOUNDS[i] = generate_r_bounds(nrings, right_edge_pixel, left_edge_pixel)
        
        filename = f'{prefix}{nrings}_{pixel_range}_{n_iterations}{suffix}' 
        f = open(filename, 'wb')
        pickle.dump(R_BOUNDS, f)
        f.close()
        path = os.getcwd()
        print('    Cached to', path + '\\' + filename)
        
        if timeit:
            print('    Time taken:', f'{(time.time() - t0):.0f}')
    
    return R_BOUNDS


def get_narrow_annuli(r_outer, dr, height, inclination, dim, n_points_per_pixel=200, timeit=True):
    '''Narrow annuli are used to speed up simulating ring images, because adding up narrow rings is faster than making up new rings. 
    This function reads in pre-generated narrow annuli if they exist. Otherwise it makes a new set of annuli with the right conditions and stores them. '''
    
    # Try stored rings
    import pickle
    prefix = 'RapidAnnuli_'
    suffix = '.python'
    found_cache = 0
    for filename in os.listdir():
        if prefix in filename:

            # Remove prefix and suffix and get parameters
            filename2 = filename.replace(prefix, '').replace(suffix, '')
            dim2, n_points_per_pixel2, height2, inclination2, r_outer2, dr2 = [float(element) for element in filename2.split('_')]
            
            # Compare parameters
            if [dim, height, inclination, dr] == [dim2, height2, inclination2, dr2] and r_outer <= r_outer2:
                print('    Found stored rings', filename2)
                found_cache = 1
                file = open(filename, 'rb')
                RINGS = pickle.load(file)
                file.close()
                break
    
    # Otherwise make rings
    if not found_cache:
        print('    Making rapid rings')
        filename = f'{prefix}{dim}_{n_points_per_pixel}_{height}_{inclination}_{r_outer}_{dr}{suffix}'
        print('    ' + filename)
        t0 = time.time()
        
        r_bounds_make = np.arange(0, r_outer+dr, dr)
        RINGS = make_all_rings(r_bounds_make, height, inclination, dim, n_points_per_pixel, kernel=None)
        
        # Store rings
        f = open(filename, 'wb')
        pickle.dump(RINGS, f)
        f.close()
        path = os.getcwd()
        print('Cached to', path + '\\' + filename)
        
        if timeit:
            print('    Time taken:', f'{(time.time() - t0):.0f}')
    
    
    # Define ring-generating functions
    def rapid_ring(r_inner, r_outer, kernel=None):
        index_inner = round(r_inner/dr)
        index_outer = round(r_outer/dr)
        assert 0 <= index_inner <= len(RINGS) and 0 <= index_outer <= len(RINGS), 'Outside rapid range!'
        ring = np.sum(RINGS[index_inner:index_outer], axis=0)
        if kernel is not None:
            ring = convolve(ring, kernel)
        return ring
    
    def rapid_rings(r_bounds, kernel=None):
        ring_images = []
        for i in range(1, len(r_bounds)):
            ring_images.append(rapid_ring(r_bounds[i-1], r_bounds[i], kernel))
        return ring_images
    
    return rapid_rings
    


## Fitting rings
def iterative_fit(target, components):
    '''Output should be identical to the matrix inversion method below but implemented in an iterative way. '''
    if type(target) is dict:
        assert type(components) is dict, 'Check input dictionary!'
        left = iterative_fit(target['left'], components['left'])
        right = iterative_fit(target['right'], components['right'])
        ratios = {'left': left[0], 'right': right[0]}
        residuals = {'left': left[1],'right': right[1]}
        return ratios, residuals
    
    n_iterations = 100
    fraction = 0.3 # How much closer each iteration gets
    
    nrings = len(components)
    all_ratios = np.zeros([n_iterations, nrings])
    residual = target.copy()
    
    for iteration in range(n_iterations):
        for iring in range(nrings-1, -1, -1):
            
            # Find ratio between image linecut and model linecut
            all_ratios[iteration, iring] = fraction * residual[iring] / components[iring, iring]
            
            # Modify redisual for next inner ring
            residual -= all_ratios[iteration, iring] * components[iring]
        
    ratios = np.sum(all_ratios, 0)
    #plot(all_ratios)
    return ratios, residual


def matrix_fit(target, components, show_matrix=False):
    '''Solves for the discretised face-on radial surface brightness given the discretised edge-on flux of modelled annuli and the observed edge-on flux of a disk. 
    EDGE-ON BRIGHTNESS OF ANNULI matrix * FACE-ON BRIGHTNESS column vector = EDGE-ON BRIGHTNESS column vector
    Input
        TARGET: a vector. Contains the observed discretised 1D flux. 
        COMPONENTS: a matrix. Contains all the modelled discretised 1D fluxes. Every column of the matrix is the discretised linecut of a ring. 
    Output
        WEIGHTS: the inverse of COMPONENTS multiplied with TARGET. 
        RESIDUALS: Always a vector of 0's since the matrix inversion method is exact. It is here to be syntactically consistent with ITERATIVE_FIT, which may have small residuals. 
    '''
    if type(target) is dict:
        assert type(components) is dict, 'Check input dictionary!'
        left = matrix_fit(target['left'], components['left'])
        right = matrix_fit(target['right'], components['right'])
        ratios = {'left': left[0], 'right': right[0]}
        residuals = {'left': left[1],'right': right[1]}
        return ratios, residuals
    
    nrings = len(components)
    matrix = np.zeros([nrings, nrings])
    
    for i in range(nrings):
        matrix[:, i] = components[i]
    
    inverted_matrix = np.linalg.inv(matrix)
    
    if False or show_matrix:
        plt.figure()
        plt.imshow(matrix)
        plt.title('Matrix')
        plt.colorbar()
        plt.show()
        
        plt.figure()
        plt.imshow(inverted_matrix)
        plt.title('Inverted matrix')
        plt.colorbar()
        plt.show()
    
    weights = inverted_matrix @ np.array(target)
    residual = np.zeros(nrings)
    return weights, residual



from .lib import *

## Brightness functions
def brightness(r):
    '''Defines the brightness of a particle as a function of distance from the star ("brightness function"). 
    Setting the brightness of each particle in the annulus to be the same, regardless of their distance from the star, is equivalent to fitting to the face-on surface *brightness* of the observation.
    Setting it to anything else is equivalent to fitting to the face-on surface *density* under the assumption that the brightness function is true.'''
    return np.ones(r.size)
    #return fancy_brightness(r)

def fancy_brightness(r):
    '''One possible brightness function. 
    This is not used in Rave by default.'''
    # Temperature
    star_luminosity = 8.7
    delta = 0.14
    A = 0.5
    #delta = 0.5
    #A = 1
    # Old A = 0.5, delta = 0.14
    # BB A = 1, delta = 0.5

    temperature = A * 278.3 * star_luminosity**0.25 * r**(-delta)

    # Blackbody (see P76 for more details)
    brightness = bnuw(wavelength, temperature)

    # Convert from Jy/sr to mJy/pixel
    brightness = brightness * (pixel*np.pi/648000)**2 * 1e3
    return brightness

def bnuw(wavelength, temperature):
    '''Blackbody function.'''
    # Give the constants for this function, where:
    # h = 6.6260755*10^{-34} Js is Planck's constant,
    # k = 1.380658*10^{-23} J/K is Boltzmann's constant,
    # c = 2.997924580*10^8 m/s is the speed of light.
    # wavelength in microns
    # temperature in K
    # Output is in Jy/sr

    k1 = 3.9728949e19 # = 2hc
    k2 = 14387.69     # = hc/k
    fact1 = k1 / wavelength**3
    fact2 = k2 / (wavelength*temperature)
    return fact1 / (np.exp(fact2) - 1)


## Rotation
def rotate_x(angle):
    '''Rotate (x, y, z) coordinates about the x axis'''
    C2 = np.zeros((3,3))

    C2[:,0] = [1, 0, 0]
    C2[:,1] = [0, np.cos(angle), np.sin(angle)]
    C2[:,2] = [0, -np.sin(angle), np.cos(angle)]
    return C2

def rotate_y(angle):
    '''Rotate (x, y, z) coordinates about the y axis'''
    C2 = np.zeros((3,3))

    C2[:,0] = [np.cos(angle), 0, -np.sin(angle)]
    C2[:,1] = [0, 1, 0]
    C2[:,2] = [np.sin(angle), 0,  np.cos(angle)]
    return C2

def rotate_z(angle):
    '''Rotate (x, y, z) coordinates about the z axis'''
    C2 = np.zeros((3,3))

    C2[:,0] = [np.cos(angle), np.sin(angle), 0]
    C2[:,1] = [-np.sin(angle), np.cos(angle), 0]
    C2[:,2] = [0, 0, 1]
    return C2


## Moving window
def moving_window(x, size=5):
    '''
    Input
        X: input array to smooth
        SIZE: size of moving window
    Returns 2 arrays of size len(x) - size + 1
        INDICES: The indices of the input array for which there is an output
        SMOOTHED: smoothed array
    '''
    x = np.array(x)
    dim = len(x)
    assert len(x.shape) == 1, 'Array to smooth must be 1D!'
    assert size <= dim, 'Window size must be at least length of array!'

    if size % 2 != 0:
        size = int(size)
        print('Adjusted window size to odd number!')

    left = size - 1
    right = dim

    shifted_x = np.zeros([size, dim + size - 1])
    for i in range(size):
        shifted_x[i, i:i+dim] = x

    smoothed = np.mean(shifted_x, 0)
    smoothed = smoothed[left:right]
    indices = np.arange(left - size//2, right - size//2) + 1
    return indices, smoothed


## Weighted sum
def weighted_sum(w, x):
    '''Sum of X weighted according to W'''
    assert len(w) == len(x)
    weighted = np.array([w[i] * x[i] for i in range(len(x))])
    return np.sum(weighted, 0)

weighted_sum_simple = weighted_sum

def weighted_sum(w, x):
    '''Sum of X weighted according to W.
    If X and/or W is a dictionary with keys 'left' and 'right', then also return a dictionary that calculates 'left' and 'right' independently.'''
    if type(w) is dict:
        if type(x) is dict:
            return {
            'left': weighted_sum_simple(w['left'], x['left']),
            'right': weighted_sum_simple(w['right'], x['right'])
            }
        else:
            return {
            'left': weighted_sum_simple(w['left'], x),
            'right': weighted_sum_simple(w['right'], x)
            }
    elif type(x) is dict:
        return {
        'left': weighted_sum_simple(w, x['left']),
        'right': weighted_sum_simple(w, x['right'])
        }
    else:
        return weighted_sum_simple(w, x)


## Dictionaries
def handlelr(function):
    '''Returns a new function such that, if there is an argument that is a dictionary, which consists of keys LEFT and RIGHT, the new function repeats the function for each of the values and outputs a dictionary with keys LEFT and RIGHT'''
    def new_function(*args, **kwargs):
        left_args = list(args)
        right_args = list(args)

        #d = None
        for i in range(len(args)):
            if type(args[i]) is dict:
                left_args[i], right_args[i] = args[i]['left'], args[i]['right']
        #assert d is not None, "Can't find dictionary to handle!"
        
        return {
        'left': function(*left_args, **kwargs),
        'right': function(*right_args, **kwargs)
        }
        
    return new_function

import matplotlib.pyplot as plt
plotlr = handlelr(plt.plot)
steplr = handlelr(plt.step)

def pltstep(x, y, *args, **kwargs):
    '''Modified version of plt.step
    X defines the boundaries of the steps
    Y defines the height of each step
    Requires SIZE(X) = SIZE(Y) + 1'''
    return plt.step(x, np.r_[y[0], y], *args, **kwargs)

def createlr(*args):
    '''Creates a dictionary with keys LEFT and RIGHT.
    If input one argument, then both keys have the same value. 
    If input two arguments, then assign them to LEFT and RIGHT respectively.'''
    assert len(args) <= 2
    if len(args) == 1:
        return {
        'left': args[0],
        'right': args[0]
        }
    else:
        return {
        'left': args[0],
        'right': args[1]
        }

def meanlr(x):
    '''Takes in a dictionary with keys LEFT and RIGHT and returns the mean of the values of the two keys'''
    return (x['left'] + x['right']) / 2

def divlr(x, y):
    '''Takes in a dictionary X with keys LEFT and RIGHT and divides the value of each key by Y'''
    return {
        'left': x['left']/y,
        'right': x['right']/y
        }

## Other functions
def yzoom(image, factor):
    '''Elongate the image along the y axis by a factor of FACTOR.'''
    image2 = zoom(image, [factor, 1])
    ydim, xdim = image2.shape
    if factor >= 1:
        cy, cx = ydim//2, xdim//2
        image2 = image2[cy - cx:cy + cx]
        assert image.shape == image2.shape
    else:
        #print(cy, cx)
        pass
    return image2


def divide(x, y):
    '''Takes in two arrays X and Y. 
    Returns Xi/Yi for each element only if Yi is not a very small number.
    Otherwise return 0 for that element. '''
    assert len(x) == len(y)
    ans = np.zeros(len(x))
    for i in range(len(x)):
        ans[i] = x[i] / y[i] if abs(y[i]) > 1e-10 else 0
    return ans


def expand(x, n_elements):
    '''If X is only a number, duplicate X into an array of length N_ELEMENTS.
    If X is already an array of length N_ELEMENTS then don't touch it. 
    Otherwise, raise an error. '''
    if type(x) in [int, float, np.float64, np.float32, np.float16, np.int64, np.int32]:
        return np.ones(n_elements) * x
    else:
        assert len(x) == n_elements
        return x

def floor(x):
    '''Floors all values in a dictionary with keys LEFT and RIGHT to 0'''
    if type(x) == dict:
        return createlr(floor(x['left']), floor(x['right']))
    assert len(x) # Make sure x has a length
    x[x < 0] = 0
    return x

def cut(image, ymax):
    '''Cut out the region of image within YMAX away from the central horizontal line'''
    y, x = image.shape
    image[ :y//2 - ymax] = 0
    image[y//2 + ymax: ] = 0
    return image

def store(variable, filename):
    '''Store VARIABLE into FILENAME with pickle'''
    import pickle
    import os
    
    f = open(filename, mode='wb')
    pickle.dump(variable, f)
    f.close()
    print('Stored at:', os.getcwd()+ '\\' + filename)


def load(filename):
    '''Load a pickled variable'''
    import pickle
    f = open(filename, mode='rb')
    variable = pickle.load(f)
    f.close()
    return variable
    

## Make kernel
def make_kernel(dim, sigma):
    '''Makes a circular Gaussian kernel with dimensions of DIM * DIM and a standard deviation of SIGMA'''
    from scipy.ndimage import gaussian_filter
    
    assert dim % 2 == 1, 'Kernel dimensions must be odd!'
    kernel = np.zeros([dim, dim])
    
    # Add a point and smooth
    kernel[dim//2, dim//2] = 1
    kernel = gaussian_filter(kernel, sigma)
    
    # Normalise
    kernel = kernel/kernel.sum()
    
    return kernel

def make_kernel2(dim, fwhm_x, fwhm_y, theta=0):
    '''Makes an elliptical Gaussian kernel with dimensions DIM * DIM and a fwhm of FWHM_X and FWHM_Y in the x and y directions respectively. 
    Then rotates the whole kernel by THETA degrees.'''
    def g(x, y, sx, sy):
        return np.exp( - (x/sx)**2/2 - (y/sy)**2/2 )
    
    # Grid coordinates
    dim2 = 2 * dim
    x = np.arange(-dim2//2, dim2//2+1)
    xx, yy = np.meshgrid(x, x)
    
    # Make kernel
    kernel = g(xx, yy, fwhm_x/2.355, fwhm_y/2.355)
    
    # Rotate
    kernel = rotate(kernel, theta)
    
    # Re-centre
    y, x = kernel.shape        
    by, bx = brightest_pixel(kernel)
    #if y % 2 == 0 and x % 2 == 0:
    #    from scipy.ndimage import shift
    #    kernel = shift(kernel, ())
    kernel = kernel[by-dim//2:by+dim//2+1, bx-dim//2:bx+dim//2+1]
    kernel /= kernel.sum()
    
    return kernel

def calculate_beam_area(fwhm_x, fwhm_y):
    '''Calculates the equivalent area of an ellitical Gaussian kernel. 
    FWHM_X can be a list of length 2, in which case FWHM_Y is ignored.'''
    if type(fwhm_x) == list:
        assert len(fwhm_x) == 2
        return calculate_beam_area(fwhm_x[0], fwhm_x[1])
    return np.pi / 4 / np.log(2) * fwhm_x * fwhm_y

def pythagoras(a, b):
    '''Calculates the length of the hypotenuse'''
    return np.sqrt(a**2 + b**2)


## Interpolate
def interpolate(RATIOS, R_BOUNDS, newpoints=1000):
    '''
    Turns a step-like function defined by only a few points into a more continuous functions by adding more sample points. 
    Input
        RATIOS: a dictionary with keys LEFT and RIGHT. Each key has an array of size N_ITERATIONS * NRINGS. 
        R_BOUNDS: a dictionary with keys LEFT and RIGHT. Each key has an array of size (NRINGS + 1). 
    Output
        RNEW: an array of length NEWPOINTS linearly spaced between the smallest and largest values of R_BOUNDS. 
        INTERPOLATED: interpolated version of RATIOS. Each key has an array of size N_ITERATIONS * NEWPOINTS.''' 
    rnew = np.linspace(R_BOUNDS[:,0].max(), R_BOUNDS[:,-1].min(), newpoints)
    assert type(RATIOS) is dict, 'RATIOS needs to be dict'
    n_iterations = RATIOS['left'].shape[0]
    INTERPOLATED = handlelr(np.zeros)([n_iterations, newpoints])
    
    for i in range(n_iterations):                
        f = interp1d(R_BOUNDS[i], np.r_[RATIOS['left'][i], RATIOS['left'][i, -1]], kind='zero')
        INTERPOLATED['left'][i] = f(rnew)
    
        f = interp1d(R_BOUNDS[i], np.r_[RATIOS['right'][i], RATIOS['right'][i, -1]], kind='zero')
        INTERPOLATED['right'][i] = f(rnew)
    return rnew, INTERPOLATED



from .lib import *
from .short_functions import *
from .height_functions import *
from .ring_functions import *
from .linecut_functions import *


class Image():
    '''Use this class to fit to an observation. '''
    
    def __init__(self, image, kernel, scale=1, distance=None, pixel=None):
        '''Create image object and specify SCALE. 
        Input
            IMAGE: 2D array representing observed image
            KERNEL: 2D array representing the PSF/beam of the observation
            SCALE: AU per pixel
            DISTANCE: distance to target in pc
            PIXEL: arcsec per pixel'''
        
        self.image = image
        self.kernel = kernel
        
        # Scale
        if distance and pixel:
            scale = distance * pixel
        
        self.scale = scale # AU/pixel
        self.distance = distance # pc
        self.pixel = pixel # arcsec/pixel
        
        # Initialise
        if image is not None:
            self.initialise()
            self.is_fake_image = False
    
    def initialise(self):
        '''Initialises IMAGE object by normalising the kernel and setting up a few variables'''
        # Dimensions
        self.ydim, self.xdim = self.image.shape # pixels
        self.cx, self.cy = self.xdim//2, self.ydim//2 # pixels
        if self.xdim % 2 != 0 or self.ydim % 2 != 0:
            print('Image dimensions not even')
        self.dim = self.xdim
        
        # Normalise kernel
        if not np.any(self.kernel == None):
            self.kernel /= self.kernel.sum()
        
        # Distance in AU
        self.R = (0.5 + np.arange(self.cx)) * self.scale # AU
        
        # Linecut
        self.linecut = make_linecut(self.image, 'full')
        
        # Other status
        if not hasattr(self, 'add_blur'):
            self.add_blur = 0
        if not hasattr(self, 'flipped'):
            self.flipped = False
    
    def add_flip(self, rotate=True):
        '''Sums the image and kernel with their repestive 180-degree rotated images'''
        if rotate:
            self.image = (self.image + rotate180(self.image)) / 2
            self.kernel = (self.kernel + rotate180(self.kernel)) / 2
        else:
            self.image = (self.image + np.fliplr(self.image)) / 2
            self.kernel = (self.kernel + np.fliplr(self.kernel)) / 2
        self.initialise()
        self.flipped = True
    
    def smooth(self, sigma=2):
        '''Smoothes the image and kernel with a circular Gaussian kernel with a standard deviation of SIGMA'''
        if sigma > 0:
            assert self.add_blur == 0, 'Already blurred!'
            self.image = blur(self.image, sigma)
            self.kernel = blur(self.kernel, sigma)
            self.kernel /= self.kernel.sum()
            self.initialise() # redo linecut
            self.add_blur = sigma
        else:
            print('Skipped smoothing')
    
    def flux_region(self, y_max, mode='cut'):
        '''Defines the region of the image that actually has flux. This is specified as the region within Y_MAX from midplane. 
        MODE can be "cut" or "specify".
            If MODE is 'SPECIFY', then only specify this region, which is subsequently used for noise calculation.
            If MODE is 'CUT', then also sets everthing outside the flux region to 0. This ignores the region with only background noise to improve S/N for subsequent fitting. '''
        self.y_max = y_max
        
        if mode == 'cut':
            self.image[ :self.cy - self.y_max] = 0
            self.image[self.cy + self.y_max: ] = 0
        
    
    def noise_by_snr(self, snr_per_beam, beam_fwhm, mode='add'):
        '''Calculates noise/pixel and noise/beam given snr/beam and beam_fwhm.
        MODE can be "add" or "specify".
            If MODE is 'SPECIFY', then only perform noise calculation.
            If MODE is 'ADD', then add this amount of noise to the image. '''
        self.snr_per_beam = snr_per_beam
        self.beam_fwhm = beam_fwhm
        
        # Number of beams required to cover the disk
        self.beam_area = calculate_beam_area(beam_fwhm, beam_fwhm)
        self.flux_pixels = self.xdim * 2 * self.y_max
        self.n_beams = self.flux_pixels / self.beam_area
        
        # Flux per beam
        total_flux = self.image.sum()
        flux_per_beam = total_flux / self.n_beams
        
        # RMS noise per beam
        self.noise_per_beam = flux_per_beam / self.snr_per_beam
        
        # RMS noise per pixel
        self.noise_per_pixel = self.noise_per_beam / np.sqrt(self.beam_area)
        
        # Add noise
        if mode == 'add':
            if self.noise_per_pixel > 0:
                noise = np.random.normal(0, self.noise_per_pixel, self.image.shape)
                self.image += noise
        
        self.kernel_noise = 0
        self.initialise()
    
    def noise_by_rms(self, image_noise, beam_fwhm=None, reference_area='pixel', mode='add'):
        '''Calculates snr/beam given (noise/pixel or noise/beam) and  beam_fwhm.
        REFERENCE_AREA can be "pixel" or "beam". 
        MODE can be "add" or "specify".
            If MODE is 'SPECIFY', then only perform noise calculation.
            If MODE is 'ADD', then add this amount of noise to the image.'''
        
        self.beam_fwhm = beam_fwhm
        
        if self.beam_fwhm:
            # Number of beams required to cover the disk
            self.beam_area = calculate_beam_area(beam_fwhm, beam_fwhm)
            self.flux_pixels = self.xdim * 2 * self.y_max
            self.n_beams = self.flux_pixels / self.beam_area
            
            # Flux per beam
            total_flux = self.image.sum()
            flux_per_beam = total_flux / self.n_beams
            
            # RMS noise per beam
            if reference_area == 'pixel':
                self.noise_per_pixel = image_noise
                self.noise_per_beam = self.noise_per_pixel * np.sqrt(self.beam_area)
            else:
                assert reference_area == 'beam'
                self.noise_per_beam = image_noise
                self.noise_per_pixel = self.noise_per_beam / np.sqrt(self.beam_area)
            
            # SNR per beam
            self.snr_per_beam = flux_per_beam / self.noise_per_beam
        
        else:
            assert reference_area == 'pixel'
            self.noise_per_pixel = image_noise
        
        # Add noise
        if mode == 'add':
            if self.noise_per_pixel > 0:
                noise = np.random.normal(0, self.noise_per_pixel, self.image.shape)
                self.image += noise
        
        self.kernel_noise = 0
        self.initialise()

    def store(self, filename):
        '''Stores IMAGE OBJECT with pickle'''
        store_image = StoreImage(self)
        store(store_image, filename)
    
    def plot(self, unit='pixel'):
        '''Plots image'''
        cy2, cx2 = self.cy + 0.5, self.cx + 0.5
        unit_label = 'pixels'
        
        if unit == 'au':
            cx2 *= self.scale
            cy2 *= self.scale
            unit_label = 'au'
        elif unit == 'beam':
            cx2 /= self.beam_fwhm
            cy2 /= self.beam_fwhm
            unit_label = 'beam FWHMs'
        plt.figure()
        plt.imshow(self.image, extent=[-cx2, cx2, -cy2, cy2], origin='lower')
        
        plt.xlabel(f'Relative RA ({unit_label})')
        plt.ylabel(f'Relative Dec ({unit_label})')
        #plt.tight_layout()
        plt.show()
    
    def make_model(self, inclination=90, heights=0, h_over_r=True, n_points_per_pixel=200, rapid=True, use_kernel=True, add_before_convolve=True, default_height=None):
        '''Makes a model of the image.
        Must be performed after radial profile is fitted so that SELF.RADIAL.PROFILE and SELF.RADIAL.RNEW exist. 
        Input
            INCLINATION: inclination to view the model image from. 0 is face-on. 90 is edge-on. 
            HEIGHTS: height of each annulus. 
                If HEIGHT is only one number, it means the height is the same everywhere. 
                If HEIGHT is NONE, the fitted height is used. Must have already obtained SELF.HEIGHT.PROFILE in this case. 
            H_OVER_R: TRUE or FALSE. If TRUE, use HEIGHTS as the aspect ratio instead of the absolute height. 
            POINTS_PER_PIXEL: number of sample points per pixel if the image was viewed edge-on for the Monte Carlo method to simulate annuli images. 
            RAPID: TRUE or FALSE. If TRUE, make model by summing up pre-generated narrow annuli. If FALSE, use Monte Carlo method. 
            USE_KERNEL: TRUE or FALSE. If True, convolve with the PSF/beam.
            ADD_BEFORE_CONVOLVE: TRUE or FALSE. If TRUE, add unconvolved narrow annuli first before convolving the final image with the PSF/beam.
            DEFAULT_HEIGHT: if HEIGHTS is NONE and SELF.HEIGHTS.PROFILE is not fitted for the entire range in r, then use DEFAULT_HEIGHT as the height everywhere not fitted to. 
            '''
        
        # Use bin_linecut as a binning function
        r_bounds_make = np.arange(0, self.cx + 1)
        profile = (self.radial.profile['left'] + self.radial.profile['right']) / 2
        weights_make = bin_linecut(profile, r_bounds_make * len(profile) / (len(r_bounds_make) - 1))
        if heights != None:
            heights_make = heights
            if h_over_r:
                rapid = False
                r = (r_bounds_make[:-1] + r_bounds_make[1:]) / 2
                heights_make = heights * r
        else:
            rapid = False
            r = self.height.rnew
            r_bounds_2 = np.arange(0, int(r[-1]) - int(r[0]) + 1)
            H = (self.height.profile['left'] + self.height.profile['right']) / 2
            H = bin_linecut(H, r_bounds_2 * len(H) / (len(r_bounds_2) - 1))
            heights_make = np.zeros(self.cx)
            heights_make[int(r[0]):int(r[-1])] = H
            if default_height == None:
                heights_make[ :int(r[0])] = self.height.starting_height
                heights_make[int(r[-1]): ] = self.height.starting_height
            elif default_height == 'interpolate':
                heights_make[ :int(r[0])] = H[0]
                heights_make[int(r[-1]): ] = H[-1]
            else:
                heights_make[ :int(r[0])] = default_height
                heights_make[int(r[-1]): ] = default_height
        
        if use_kernel:
            kernel = self.kernel
        else:
            kernel = None
        
        self.model = MakeImage(r_bounds_make, weights_make, heights_make, inclination, self.xdim, n_points_per_pixel, kernel, scale=self.scale, rapid=rapid, add_before_convolve=add_before_convolve)
    
    def plot_compare(self, cut_height=10, unit='pixel'):
        '''Makes a plot to compare the full 1D flux and midplane flux between the observation and best-fit model.'''
        if unit == 'pixel':
            r = np.arange(self.cx)
            unit_label = 'pixels'
        elif unit == True or unit == 'au':
            r = np.arange(self.cx) * self.scale
            unit_label = 'au'
        elif unit == 'beam':
            assert self.beam_fwhm != None
            r = np.arange(self.cx) / self.beam_fwhm
            unit_label = 'beam FWHMs'
        
        plt.figure()
        plt.plot(r, meanlr(make_linecut(self.image, 'full')), '--', label='Data: full flux')
        plt.plot(r, meanlr(make_linecut(self.model.image, 'full')), label='Model: full flux')
        plt.plot(r, meanlr(make_linecut(self.image, cut_height)), '--', label='Data: midplane flux')
        plt.plot(r, meanlr(make_linecut(self.model.image, cut_height)), label='Model: midplane flux')
        plt.xlabel(f'Radial distance ({unit_label})')
        plt.ylabel('1D flux')
        plt.legend(frameon=False)
        plt.tight_layout()
        plt.show()
    
    def fit_inclination(self, cut_height=10, inc_range=[0, 90], guess=58, delta=1, unit='pixel'):
        '''Finds a lower bound for the inclination.
        Must have already fitted SELF.RADIAL.PROFILE.'''
        
        if unit == 'pixel':
            r = np.arange(self.cx)
            unit_label = 'pixels'
        elif use_au == True:
            r = np.arange(self.cx) * self.image.scale
            unit_label = 'au'
        elif use_au == 'beam':
            assert self.beam_fwhm != None
            r = np.arange(self.cx) / self.beam_fwhm
            unit_label = 'beam FWHMs'
        
        lim = make_linecut(self.image, cut_height)
        lim = meanlr(lim)
        plt.figure()
        plt.plot(r, lim, label='Data')
        
        while inc_range[1] - inc_range[0] > 2:
            self.make_model(inclination=guess, heights=0, add_before_convolve=True)
            lmod = make_linecut(self.model.image, cut_height)
            lmod = meanlr(lmod)
            plt.plot(r, lmod, label=guess)
            
            if np.sum(lim > lmod + delta) > 0:
                inc_range[0] = guess
            else:
                inc_range[1] = guess
            guess = int(np.round(np.mean(inc_range)))
        
        print('Lower bound for inclination:', inc_range)
        self.inclination = inc_range[1]
        
        plt.xlabel(f'Radial distance ({unit_label})')
        plt.ylabel('Midplane flux')
        plt.legend(frameon=False)
        plt.tight_layout()
        plt.show()
    
    def estimate_inclination(self, cut_height=10, inc_range=[78, 90], di=2, unit='pixel'):
        '''Finds a lower bound for the inclination.
        Must have already fitted SELF.RADIAL.PROFILE.'''
        
        if unit == 'pixel':
            r = np.arange(self.cx)
            unit_label = 'pixels'
        elif use_au == True:
            r = np.arange(self.cx) * self.image.scale
            unit_label = 'au'
        elif use_au == 'beam':
            assert self.beam_fwhm != None
            r = np.arange(self.cx) / self.beam_fwhm
            unit_label = 'beam FWHMs'
        
        lim = make_linecut(self.image, cut_height)
        lim = meanlr(lim)
        plt.figure()
        plt.plot(r, lim, label='Data')
        
        inc_vals = np.arange(inc_range[0], inc_range[1]+di, di)
        
        for inc in inc_vals:
            self.make_model(inclination=inc, heights=0, add_before_convolve=True)
            lmod = make_linecut(self.model.image, cut_height)
            lmod = meanlr(lmod)
            plt.plot(r, lmod, label=inc)
        
        plt.xlabel(f'Radial distance ({unit_label})')
        plt.ylabel('Midplane flux')
        plt.legend(frameon=False)
        plt.tight_layout()
        plt.show()


class MakeImage(Image):
    '''Use this class to make up an image for testing'''
    
    def __init__(self, r_bounds_make, weights_make, heights_make, inclination_make, dim, n_points_per_pixel, kernel, scale=1, rapid=False, add_before_convolve=True, verbose=True):
        '''Specify parameters used to make up the image.
        All units are in pixels. 
        Input:
            R_BOUNDS_MAKE: boundaries defining the annuli. 
            WEIGHTS_MAKE: face-on surface brightness of each annulus. 
            HEIGHTS_MAKE: standard deviation of the Gaussian distribution of  the vertical height of each annulus. 
            INCLINATION_MAKE: inclination to view the image from. 0 is face-on. 90 is edge-on. 
            DIM: dimensions of the image. Output image is always square. 
            N_POINTS_PER_PIXEL: how many points to add to the image per pixel if the ring were to be viewed face-on. The more points you add, the higher the accuracy of the image, but the slower the code will run. THIS MUST BE CONSISTENT WITH R IF R IS GIVEN. 
            KERNEL: the kernel to concolve the image with. None means don't convolve with anything. 
            SCALE: AU per pixel.  
            RAPID: TRUE or FALSE. If TRUE, make model by summing up pre-generated narrow annuli. If FALSE, use Monte Carlo method. 
            ADD_BEFORE_CONVOLVE: TRUE or FALSE. If TRUE, add unconvolved narrow annuli first before convolving the final image with the PSF/beam.
        '''
            
        self.r_bounds_make = r_bounds_make
        self.weights_make = weights_make
        self.heights_make = heights_make
        self.inclination_make = inclination_make
        self.dim = dim
        self.n_points_per_pixel = n_points_per_pixel
        self.scale = scale
        
        # Make image
        if verbose:
            print('Making image')
        if not rapid:
            t0 = time.time()
            if not add_before_convolve:
                rings_make = make_all_rings(r_bounds_make, heights_make, inclination_make, dim, n_points_per_pixel, kernel, verbose=verbose)
            else:
                rings_make = make_all_rings(r_bounds_make, heights_make, inclination_make, dim, n_points_per_pixel, None, verbose=verbose)
            if verbose:
                print('Time taken:', f'{(time.time() - t0):.0f}')
        else:
            self.get_rapid_annuli()
            if not add_before_convolve:
                rings_make = self.rapid_rings(r_bounds_make, kernel)
            else:
                rings_make = self.rapid_rings(r_bounds_make, None)
        
        image = weighted_sum(weights_make, rings_make)
        if add_before_convolve and not np.any(kernel == None):
            image = convolve(image, kernel)
        
        self.image = image
        self.kernel = kernel
        self.rings_make = rings_make
        
        # Initialise
        self.initialise()
        self.is_fake_image = True
    
    def get_rapid_annuli(self, dr=0.1):
        '''Load the rapid annuli used to generate the fake image.'''
        print('\n----- Make image annuli -----')
        
        self.rapid_rings = get_narrow_annuli(self.dim//2, dr, self.heights_make, self.inclination_make, self.dim, self.n_points_per_pixel)
    

class LoadImage(MakeImage):
    '''Use this class to load a stored Image Object'''

    def __init__(self, filename, scale=None, kernel=None, fitted=False):
        '''Specify SCALE and KERNEL. 
        Input
            FILENAME: the location and name of the stored image to load. 
            SCALE: AU per pixel. 
            KERNEL: PSF/beam of the observation. 
            FITTED: whether or not this image object contains fitted profiles. '''
        
        loaded_image = load(filename)
        self.image = loaded_image.image
        self.kernel = loaded_image.kernel
        
        # Store variables
        if not fitted:
            self.r_bounds_make = loaded_image.r_bounds_make
            self.weights_make = loaded_image.weights_make
            self.heights_make = loaded_image.heights_make
            self.inclination_make = loaded_image.inclination_make
        self.scale = scale if scale else loaded_image.scale
        
        # Convolve
        if np.any(kernel != None):
            assert np.all(loaded_image.kernel == None)
            self.kernel = kernel / kernel.sum()
            self.image = convolve(self.image, self.kernel)
        
        # Initialise
        self.initialise()
        self.is_fake_image = True


class Empty():
    '''Empty class'''
    pass


class StoreImage():
    '''Use this class to create an image object that can be stored.'''
    
    def __init__(self, image_object):
        '''Creates a new image object with the same attributes except ones that can't be pickled'''
        for x in image.__dict__.keys():
            if x == 'radial':
                self.radial = Empty()
                for y in image.radial.__dict__.keys():
                    if y not in ['rapid_rings', 'image']:
                        val = getattr(image.radial, y)
                        setattr(self.radial, y, val)
            
            elif x == 'height':
                self.height = Empty()
                for z in image.height.__dict__.keys():
                    if z not in ['RAPID_RINGS', 'image']:
                        val = getattr(image.height, z)
                        setattr(self.height, z, val)
            
            else:
                if x not in ['model', 'rings_make']:
                    val = getattr(image, x)
                    setattr(self, x, val)


class RadialProfile():
    '''Radial Profile Fitter class. Here "Radial Profile" = "surface brightness profile".'''
    
    def __init__(self, image):
        '''Links the Image object with the Radial Profile Fitter obeject'''
        image.radial = self
        self.image = image
    
    def quick_fit(self, nrings=7, r_outer=None, fit_method='matrix', r_bounds='even', ring_method='rapid'):
        '''Does a quick fit to the image with only one iteration. 
        Input
            NRINGS: number of annuli to use. 
            R_OUTER: fit up to this radial location. 
            FIT_METHOD: 'matrix' or 'iterative'.
            R_BOUNDS: boundaries defining the annuli. If 'even', then evenly spaces the annuli. 
            RING_METHOD: if 'rapid' then makes rings by summing up narrow annuli. Otherwise makes up rings from scratch using Monte Carlo. '''
        print('\n----- Radial Quick Fit -----')
        
        t0 = time.time()
        
        if r_bounds == 'even':
            if r_outer == None:
                if hasattr(self, 'r_outer'):
                    r_outer = self.r_outer
                else:
                    r_outer = self.image.cx
            r_bounds = np.linspace(0, r_outer, nrings+1)
        
        if fit_method == 'matrix':
            function = matrix_fit
        else:
            function = iterative_fit
        
        if ring_method == 'rapid':
            rings = self.rapid_rings(r_bounds, self.image.kernel)
        else:
            print('    Making fitting annuli')
            t0 = time.time()
            rings = make_all_rings(r_bounds, heights=10, inclination=90, dim=self.image.xdim, n_points_per_pixel=200, kernel=self.image.kernel)
            print('    Time taken:', f'{(time.time() - t0):.0f}')
        
        self.quick_abrl = bin_rings(rings, r_bounds, 'full')
        L = bin_linecut(self.image.linecut, r_bounds)
        self.quick_profile, _ = function(L, self.quick_abrl)
        self.quick_r_bounds = r_bounds
        self.quick_nrings = nrings
        
        print('Time taken:', f'{(time.time() - t0):.0f}')
    
    def quick_plot(self, unit='pixel', average=False):
        '''Makes a quick plot of the QUICK_FIT. 
        UNIT can be 'pixel', 'au' or 'beam'. '''
        
        if unit == 'pixel':
            r1 = self.quick_r_bounds
            r2 = self.image.r_bounds_make
            unit_label = 'pixels'
        if unit == 'au':
            r1 = self.quick_r_bounds * self.image.scale
            r2 = self.image.r_bounds_make * self.image.scale
            unit_label = 'au'
        elif unit == 'beam':
            r1 = self.quick_r_bounds / self.image.beam_fwhm
            r2 = self.image.r_bounds_make / self.image.beam_fwhm
            unit_label = 'beam FWHMs'
            
        plt.figure()
        if not average:
            pltstep(r1, self.quick_profile['left'], label='Left')
            pltstep(r1, self.quick_profile['right'], label='Right')
        else:
            pltstep(r1, meanlr(self.quick_profile), label='Fitted')
        if self.image.is_fake_image:
            pltstep(r2, self.image.weights_make, linestyle='--', label='True', color=colours[3], alpha=0.8)
        plt.legend(frameon=False)
        plt.title('Quick radial profile')
        plt.xlabel(f'Radial distance ({unit_label})')
        plt.ylabel('Surface brightness (Jy/beam)')
        plt.tight_layout()
        plt.show()
    
    def get_rapid_annuli(self, r_outer, dr=0.1, height=10, inclination=90, points_per_pixel=200):
        '''Gets narrow annuli with the required properties. 
        Input:
            R_OUTER: outermost radial location with flux to fit to. 
            DR: width of each narrow annulus. 
            HEIGHT: height of each narrow annulus. 
            INCLINATION: inclination of each narrow annulus.'''
        print('\n----- Radial annuli -----')
        
        self.r_outer = r_outer
        self.dr = dr
        self.height = height
        self.inclination = inclination
        self.points_per_pixel = points_per_pixel
        
        # Get function that makes rings rapidly
        self.rapid_rings = get_narrow_annuli(r_outer, dr, height, inclination, self.image.xdim, points_per_pixel)
    
    def fit(self, nrings, n_iterations=100, extra_noise=0, random=True, verbose=True, fit_star=False, floor_to_0=True):
        '''Fits the radial surface brightness profile.
        Input
            NRINGS: number of annuli to use. 
            N_ITERATIONS: how many sets of fits to perform. Each set has a different set of annuli boundaries. 
            EXTRA_NOISE: how much noise to add to the best-fit model to repeat the fitting procedure on. If 0, the code will skip this step. 
            RANDOM: whether or not to randomly pick annuli boundaries if there are more sets of stored annuli boundaries than needed. 
            FIT_STAR: if True, then adds an annulus that occupies the first pixel to fit to the flux of the star. 
            FLOOR_TO_0: whether or not to set all fitted values below 0 to 0. '''
        
        ## Preparation
        print('\n----- Radial Fit -----')
        
        # Set up attributes
        self.nrings = nrings
        self.n_iterations = n_iterations
        self.extra_noise = extra_noise
        self.fit_star = fit_star
        
        # Set up variables
        dr = self.dr
        kernel = self.image.kernel
        image_linecut = self.image.linecut
        ndim = nrings + fit_star
        
        # Get annuli boundaries
        if not fit_star:
            self.R_BOUNDS = get_r_bounds(self.nrings, self.r_outer, n_iterations)
        else:
            "Include an additional ring with radius = 1 pixel"
            R_BOUNDS = get_r_bounds(self.nrings, [1, self.r_outer], n_iterations, use_flux_range=True)
            R_BOUNDS2 = np.zeros([len(R_BOUNDS), ndim+1])
            for i in range(len(R_BOUNDS)):
                R_BOUNDS2[i] = np.r_[0, R_BOUNDS[i]]
            self.R_BOUNDS = R_BOUNDS2
        
        # Set up storing variables
        self.R_BOUNDS = np.round(self.R_BOUNDS / dr) * dr
        self.MTXRATIOS = handlelr(np.zeros)([n_iterations, ndim])
        "fitted with extra noise"
        self.RATIOS_2 = handlelr(np.zeros)([n_iterations, ndim]) 
        
        # Pick subset of boundaries to use 
        if random:
            indices = np.random.choice(len(self.R_BOUNDS), n_iterations, replace=False)
            self.R_BOUNDS = self.R_BOUNDS[indices]
        else:
            self.R_BOUNDS = self.R_BOUNDS[0:n_iterations]
        
        # Time it
        t0 = time.time()
        
        ## Fit for each set of annuli boundaries
        for i in range(n_iterations):
            
            # Print status
            if verbose: print(i, end=' ')
            
            # Make annuli
            r_bounds = self.R_BOUNDS[i]
            rings = self.rapid_rings(r_bounds, kernel)
            abrl = bin_rings(rings, r_bounds, 'full')
            
            # Fit with matrix and iterative method
            L = bin_linecut(image_linecut, r_bounds)
            mtxratios, _ = matrix_fit(L, abrl)
        
            # Store values
            self.MTXRATIOS['left'][i], self.MTXRATIOS['right'][i] = mtxratios['left'], mtxratios['right']
            
        # Interpolate points
        self.rnew, INTERPOLATED = interpolate(self.MTXRATIOS, self.R_BOUNDS)
        
        ## Make noiseless model
        # *** Store profile ***
        self.profile = handlelr(np.median)(INTERPOLATED, axis=0)
        
        # Generate model
        if extra_noise:
            self.image.make_model(inclination=90, heights=0)
            noiseless_model = self.image.model.image
            
        ## Fit again with noise
        for i in range(n_iterations):
            
            # Don't fit with extra noise if no extra noise specified
            if not extra_noise: continue
            
            # Print status
            if verbose: print('n', i, end=' ')
            
            # Make annuli
            r_bounds = self.R_BOUNDS[i]
            rings = self.rapid_rings(r_bounds, kernel)
            abrl = bin_rings(rings, r_bounds, 'full')
        
            # Add noise to model
            noise = np.random.normal(0, extra_noise, self.image.image.shape)
            if self.image.add_blur:
                noise = blur(noise, self.image.add_blur)
            noisy_model = noiseless_model + noise
            noisy_model = cut(noisy_model, self.image.y_max)
            if self.image.flipped:
                noisy_model = (noisy_model + rotate180(noisy_model)) / 2
            "Make linecut"
            noisy_linecut = make_linecut(noisy_model, 'full')
            
            # Fit with matrix method
            L2 = bin_linecut(noisy_linecut, r_bounds)
            mtxratios2, _ = matrix_fit(L2, abrl)
        
            # Store values
            self.RATIOS_2['left'][i], self.RATIOS_2['right'][i] = mtxratios2['left'], mtxratios2['right']
        
        if not extra_noise:
            self.RATIOS_2 = self.MTXRATIOS
            
        # Interpolate points        
        self.rnew, INTERPOLATED2 = interpolate(self.RATIOS_2, self.R_BOUNDS)
        
        ## Calculate percentiles
        quantile_range = 0.34
        
        # Uncertainty without noise [first shade of uncertainty]
        self.profile_up = handlelr(np.quantile)(INTERPOLATED, 0.5 + quantile_range, axis=0)
        self.profile_down = handlelr(np.quantile)(INTERPOLATED, 0.5 - quantile_range, axis=0)
        
        # Uncertainty with noise [not used]
        self.profile2 = handlelr(np.median)(INTERPOLATED2, axis=0)
        
        # Uncertainty without noise [second shade of uncertainty]
        self.profile_up2 = handlelr(np.quantile)(INTERPOLATED2, 0.5 + quantile_range, axis=0)
        self.profile_down2 = handlelr(np.quantile)(INTERPOLATED2, 0.5 - quantile_range, axis=0)
        
        # Set negative values to 0
        if floor_to_0:
            self.profile = floor(self.profile)
            self.profile_up = floor(self.profile_up)
            self.profile_down = floor(self.profile_down)
            
            self.profile = floor(self.profile)
            self.profile_up2 = floor(self.profile_up2)
            self.profile_down2 = floor(self.profile_down2)
        
        print('Time taken:', f'{(time.time() - t0):.0f}')
        
    def plot(self, smooth=0, average=True, factor=1, use_au=True, floor_to_0=True):
        '''Plots the fitted radial profile. 
        Input
            SMOOTH: how much to smooth the fitted curves by. Recommend trying 101 if you want to smooth. 
            AVERAGE: whether or not to average the fits to the left and right halves of the image. In most cases you probably have averaged the left and right halves of the image already, so the two fits should be the same. Recommend using TRUE.
            FACTOR: how much to vertical scale the fit by. Used to avoid numbers that are too small or big on the axis labels when they the absolute value of the values doesn't matter. 
            USE_AU: this keyword is identical to UNIT in other methods. Can be 'pixel', 'au' or 'beam'.
            FLOOR_TO_0: whether or not to set everything below 0 to 0 when plotting. '''
        
        if smooth:
            from scipy.signal import savgol_filter
            window_length = smooth
            polyorder = 3
            
            yup = handlelr(savgol_filter)(self.profile_up, window_length, polyorder)
            ymedian = handlelr(savgol_filter)(self.profile, window_length, polyorder)
            ydown = handlelr(savgol_filter)(self.profile_down, window_length, polyorder)
            
            yup2 = handlelr(savgol_filter)(self.profile_up2, window_length, polyorder)
            ymedian2 = handlelr(savgol_filter)(self.profile2, window_length, polyorder)
            ydown2 = handlelr(savgol_filter)(self.profile_down2, window_length, polyorder)
            
        else:
            yup = self.profile_up
            ymedian = self.profile
            ydown = self.profile_down
            
            yup2 = self.profile_up2
            ymedian2 = self.profile2
            ydown2 = self.profile_down2
        
        yup = divlr(yup, 1/factor)
        ymedian = divlr(ymedian, 1/factor)
        ydown = divlr(ydown, 1/factor)
        
        yup2 = divlr(yup2, 1/factor)
        ymedian2 = divlr(ymedian2, 1/factor)
        ydown2 = divlr(ydown2, 1/factor)
        
        if not use_au:
            r = self.rnew
            if self.image.is_fake_image:
                r2 = self.image.r_bounds_make
            unit_label = 'pixels'
        elif use_au == True:
            r = self.rnew * self.image.scale
            if self.image.is_fake_image:
                r2 = self.image.r_bounds_make * self.image.scale
            unit_label = 'au'
        elif use_au == 'beam':
            assert self.image.beam_fwhm != None
            r = self.rnew / self.image.beam_fwhm
            if self.image.is_fake_image:
                r2 = self.image.r_bounds_make / self.image.beam_fwhm
            unit_label = 'beam FWHMs'
        
        # floor to 0
        if floor_to_0:
            ymedian = floor(ymedian)
            yup = floor(yup)
            ydown = floor(ydown)
            
            ymedian2 = floor(ymedian2)
            yup2 = floor(yup2)
            ydown2 = floor(ydown2)
        
        plt.figure()
        if not average:
            plt.plot(r, ymedian['left'], alpha=0.8, linewidth=2, label='Left')
            plt.plot(r, ymedian['right'], alpha=0.8, linewidth=2, label='Right')
            handlelr(plt.fill_between)(r, yup, ydown, alpha=0.5, linewidth=2)
        else:
            plt.plot(r, meanlr(ymedian), alpha=0.8, linewidth=2, label='Fitted from obs.')
            plt.fill_between(r, meanlr(yup), meanlr(ydown), alpha=0.5, linewidth=0)
            if self.extra_noise:
                plt.plot(r, meanlr(ymedian2), alpha=0.8, linewidth=2, label='Fitted from model')
                plt.fill_between(r, meanlr(yup2), meanlr(ydown2), color='C1', alpha=0.2, linewidth=0)
        #plt.ylim([None, np.max(meanlr(yup)[100:])*1.2])
        
        if self.image.is_fake_image:
            pltstep(r2, self.image.weights_make, color=colours[3], alpha=0.8, linestyle='--', label='True')
        
        if not average or self.image.is_fake_image:
            plt.legend(frameon=False)
        plt.title(f'{self.nrings} annuli, {self.image.noise_per_pixel/self.image.image.max()*100:.0f}% noise ({self.n_iterations} iterations)')
        plt.xlabel(f'Radial distance ({unit_label})')
        plt.ylabel(r'Surface brightness ($\mu$Jy/beam)')
        plt.tight_layout()
        plt.show()
    
        
class HeightProfile():
    '''Height Profile Fitter class. '''
    
    def __init__(self, image):
        '''Links the Image object with the Height Profile Fitter obeject'''
        image.height = self
        self.image = image
    
    def get_rapid_annuli(self, r_outer, dr=0.1, hrange=[10, 30], dh=3, inclination=90):
        '''Gets narrow annuli with the required properties. 
        Input:
            R_OUTER: outermost radial location with flux to fit to. 
            DR: width of each narrow annulus. 
            HRANGE: minimum and maximum height of annuli to load.
            DH: spacing in heights between each set of narrow annuli to load. 
            INCLINATION: inclination of each narrow annulus.'''
        print('\n----- Height annuli -----')
        
        self.r_outer = r_outer
        self.dr = dr
        self.points_per_pixel = self.image.radial.points_per_pixel
        
        self.interp_heights = np.arange(hrange[0], hrange[-1] + dh, dh)
        self.dh = dh
        self.inclination = inclination
        
        # Storing parameter
        self.RAPID_RINGS = dict()
        
        for height in self.interp_heights:
            self.RAPID_RINGS[height] = get_narrow_annuli(self.r_outer, self.dr, height, inclination, self.image.xdim, self.points_per_pixel)
        
    def quick_fit(self, cut_height, starting_height):
        '''Does a quick fit to the image with only one iteration. Must have done a QUICK_FIT to the radial profile already. 
        Input
            CUT_HEIGHT: distance from the midplane defining the midplane region.
            STARTING_HEIGHT: intial height of the annuli to initiate the algorithm.  '''
        print('\n----- Height Quick Fit -----')
        
        r_bounds = self.image.radial.quick_r_bounds
        mtxratios = self.image.radial.quick_profile
        
        t0 = time.time()
        
        # Get calibrators and generators
        height_calibrators, part_generators = make_generators(r_bounds, self.interp_heights, self.image.xdim, self.RAPID_RINGS, self.image.kernel, cut_height)
        
        # Fit height
        self.image.part = make_linecut(self.image.image, cut_height)
        self.lim = bin_linecut(self.image.part, r_bounds)
        H, lmod = fit_height(self.lim, mtxratios, r_bounds, height_calibrators, part_generators, starting_height, self.image.xdim)
            
        self.quick_profile = createlr(H['left'][-1, :], H['right'][-1, :])
        self.quick_lmod = createlr(lmod['left'][-1, :], lmod['right'][-1, :])
        self.H = H
        self.lmod = lmod
        self.quick_r_bounds = r_bounds
        self.height_calibrators = height_calibrators
        self.part_generators = part_generators
        print('Time taken:', f'{(time.time() - t0):.0f}')
    
    def quick_plot(self, unit='pixel'):
        '''Makes a quick plot of the QUICK_FIT. 
        UNIT can be 'pixel', 'au' or 'beam'. '''
        
        if unit == 'pixel':
            r1 = self.quick_r_bounds
            r2 = self.image.r_bounds_make
            unit_label = 'pixels'
        if unit == 'au':
            r1 = self.quick_r_bounds * self.image.scale
            r2 = self.image.r_bounds_make * self.image.scale
            unit_label = 'au'
        elif unit == 'beam':
            r1 = self.quick_r_bounds / self.image.beam_fwhm
            r2 = self.image.r_bounds_make / self.image.beam_fwhm
            unit_label = 'beam FWHMs'
        
        plt.figure()
        pltstep(r1, self.quick_profile['left'], label='Left')
        pltstep(r1, self.quick_profile['right'], label='Right')
        plt.legend(frameon=False)
        if self.image.is_fake_image:
            pltstep(r2, self.image.heights_make, linestyle='--', label='True')
        plt.legend(frameon=False)
        plt.title('Quick height profile')
        plt.xlabel(f'Radial distance ({unit_label})')
        plt.ylabel(f'Height ({unit_label})')
        plt.tight_layout()
        
        
        plt.figure()
        pltstep(r1, self.quick_lmod['left'], label='Left')
        pltstep(r1, self.quick_lmod['right'], label='Right')
        pltstep(r1, self.lim['right'], label='Observed')
        plt.legend(frameon=False)
        plt.title('Quick model partial flux')
        plt.xlabel(f'Radial distance ({unit_label})')
        plt.ylabel('Partial flux')
        plt.tight_layout()
        
    
    def fit(self, nrings, cut_height, starting_height, n_iterations=100, extra_noise=0, flux_range=None, random=True, verbose=True, remove_default=True):
        '''Fits the radial height profile.
        Input
            NRINGS: number of annuli to use. 
            CUT_HEIGHT: distance from the midplane defining the midplane region.
            STARTING_HEIGHT: intial height of the annuli to initiate the algorithm.  
            N_ITERATIONS: how many sets of fits to perform. Each set has a different set of annuli boundaries. 
            EXTRA_NOISE: how much noise to add to the best-fit model to repeat the fitting procedure on. If 0, the code will skip this step. 
            FLUX_RANGE: the range in radial distance from the star that actually has substantial flux. Other regions can't be fitted to since there is not enough flux to generate a robust fit. The heights of those regions is assumed to be STARTING_HEIGHT. 
            RANDOM: whether or not to randomly pick annuli boundaries if there are more sets of stored annuli boundaries than needed. 
            REMOVE_DEFAULT: whether or not to remove the regions without substantial flux by creating a model of those regions. Recommending using TRUE.'''
        
        ## Preparation
        print('\n----- Height Fit -----')
        
        # Set up attributes
        self.cut_height = cut_height
        self.starting_height = starting_height
        self.n_iterations = n_iterations
        self.extra_noise = extra_noise
        self.nrings = nrings
        
        "CUT_HEIGHT is used to to define the partial linecut"
        self.image.part = make_linecut(self.image.image, cut_height)
        
        "Height fitting is only done to radial locations where the fitted surface brightness is within FLUX_RANGE"
        if flux_range == None:
            self.flux_range = [0, self.r_outer]
        else:
            self.flux_range = flux_range
        assert n_iterations <= self.image.radial.n_iterations
        
        # Time it
        t0 = time.time()
        
        ## Default model
        "The 'default model' is used to remove the midplane flux contribution from regions where the surface brightness is low"
        "Fitting directly to their height will influence regions where there actually is flux"
        
        if remove_default and (flux_range[0] != 0 or flux_range[1] != self.r_outer):
            
            print('Removing default model from midplane flux')
            
            # Radial bins for making default model
            rb = np.arange(self.r_outer+1)
            r = (rb[1:] + rb[:-1]) / 2
            
            # Get surface brightness
            rnew = self.image.radial.rnew
            delta = rnew[1] - rnew[0]
            radial_profile = meanlr(self.image.radial.profile)
            radial_profile[radial_profile < 0] = 0
            w = bin_linecut(radial_profile, rb/delta)
            
            # Remove contributions from regions about to be fitted to
            w[flux_range[0]:flux_range[1]] = 0
            "Using starting height as default model height"
            h = starting_height
            default_model = MakeImage(rb, w, h, inclination_make=self.inclination, dim=self.image.dim, n_points_per_pixel=self.points_per_pixel, kernel=self.image.kernel, scale=1, rapid=False, add_before_convolve=True)
            self.default_model = default_model
            
            # Remove midplane flux contribution from default model
            default_part = make_linecut(default_model.image, cut_height)
            self.image.part['left'] -= default_part['left']
            self.image.part['right'] -= default_part['right']
        
        ## More preparation
        # Get annuli boundaries
        self.R_BOUNDS = get_r_bounds(self.nrings, self.flux_range, n_iterations, use_flux_range=True)
        if random:
            indices = np.random.choice(len(self.R_BOUNDS), n_iterations, replace=False)
            self.R_BOUNDS = self.R_BOUNDS[indices]
        else:
            self.R_BOUNDS = self.R_BOUNDS[0:n_iterations]
        
        # Storing variables
        self.HEIGHT = handlelr(np.zeros)([n_iterations, nrings])
        self.HEIGHT_2 = handlelr(np.zeros)([n_iterations, nrings])
        
        ## Fit height for each set of annuli boundaries
        for i in range(n_iterations):
            
            # Print status
            if verbose >= 1: print(i, end=' ')
            
            # Get annuli boundaries
            r_bounds = self.R_BOUNDS[i]
            
            # Get HEIGHT_CALIBRATORS and PART_GENERATORS for all annuli
            "HEIGHT_CALIBRATORS: mapping between h and midplane flux"
            "PART_GENERATORS: generates midplane flux given height"
            height_calibrators, part_generators, intp_range = make_generators(r_bounds, self.interp_heights, self.image.xdim, self.RAPID_RINGS, self.image.kernel, cut_height, testing=True)
            
            "Store these functions for testing"
            self.height_calibrators = height_calibrators
            self.part_generators = part_generators
            self.intp_range = intp_range
            
            # Fit height
            lim = bin_linecut(self.image.part, r_bounds)
            H, lmod = fit_height2(lim, self.image.radial.profile, self.image.radial.rnew, r_bounds, height_calibrators, part_generators, starting_height, self.image.xdim, intp_range)
            
            # Store values
            self.HEIGHT['left'][i, :], self.HEIGHT['right'][i, :] = H['left'][-1, :], H['right'][-1, :]
        
        # Interpolate height
        self.rnew, HEIGHT_INT = interpolate(self.HEIGHT, self.R_BOUNDS)
        
        ## Make noiseless model
        # *** Store profile ***
        self.profile = handlelr(np.median)(HEIGHT_INT, axis=0)
        
        # Generate model
        if extra_noise:
            self.image.make_model(inclination=self.inclination, heights=None, h_over_r=False, n_points_per_pixel=200, rapid=True, use_kernel=True, default_height=starting_height)

            noiseless_model = self.image.model.image
        
        ## Fit again with noise
        for i in range(n_iterations):
            
            # Don't fit with extra noise if no extra noise specified
            if not extra_noise: continue
            
            # Print status
            if verbose: print('n', i, end=' ')
            
            # Get annuli boundaries
            r_bounds = self.R_BOUNDS[i]
            
            # Get HEIGHT_CALIBRATORS and PART_GENERATORS for all annuli
            "HEIGHT_CALIBRATORS: mapping between h and midplane flux"
            "PART_GENERATORS: generates midplane flux given height"
            height_calibrators, part_generators, intp_range = make_generators(r_bounds, self.interp_heights, self.image.xdim, self.RAPID_RINGS, self.image.kernel, cut_height, testing=True)
            
            "Store these functions for testing"
            self.height_calibrators = height_calibrators
            self.part_generators = part_generators
            self.intp_range = intp_range
            
            # Add noise to model
            noise = np.random.normal(0, extra_noise, self.image.image.shape)
            if self.image.add_blur:
                noise = blur(noise, self.image.add_blur)
            noisy_model = noiseless_model + noise
            noisy_model = cut(noisy_model, self.image.y_max)
            if self.image.flipped:
                noisy_model = (noisy_model + rotate180(noisy_model)) / 2
            
            "Make linecut"
            #plot(noisy_model)
            part2 = make_linecut(noisy_model, cut_height)
            if remove_default and (flux_range[0] != 0 or flux_range[1] != self.r_outer):
                part2['left'] -= default_part['left']
                part2['right'] -= default_part['right']
            lim2 = bin_linecut(part2, r_bounds)
            
            # Fit height
            H2, lmod2 = fit_height2(lim2, self.image.radial.profile, self.image.radial.rnew, r_bounds, height_calibrators, part_generators, starting_height, self.image.xdim, intp_range)
            
            # Store values
            self.HEIGHT_2['left'][i, :], self.HEIGHT_2['right'][i, :] = H2['left'][-1, :], H2['right'][-1, :]
        
        # Interpolate height
        self.rnew, HEIGHT_2_INT = interpolate(self.HEIGHT_2, self.R_BOUNDS)
        
        if not extra_noise:
            HEIGHT_2_INT = HEIGHT_INT
        
        ## Calculate percentiles
        quantile_range = 0.34
        self.profile_up = handlelr(np.quantile)(HEIGHT_INT, 0.5 + quantile_range, axis=0)
        self.profile_down = handlelr(np.quantile)(HEIGHT_INT, 0.5 - quantile_range, axis=0)
        
        self.profile2 = handlelr(np.median)(HEIGHT_2_INT, axis=0)
        self.profile_up2 = handlelr(np.quantile)(HEIGHT_2_INT, 0.5 + quantile_range, axis=0)
        self.profile_down2 = handlelr(np.quantile)(HEIGHT_2_INT, 0.5 - quantile_range, axis=0)
         
        print(f'Time taken: {(time.time() - t0):.0f}')

    def plot(self, smooth=0, F_lim=0, average=True, use_au=False, h_over_r=False):
        '''Plots the fitted radial profile. 
        Input
            SMOOTH: how much to smooth the fitted curves by. Recommend trying 101 if you want to smooth. 
            F_LIM: only plots the heights for regions with flux above F_LIM. Recommending setting to 0 if using REMOVE_DEFAULT when fitting. 
            AVERAGE: whether or not to average the fits to the left and right halves of the image. In most cases you probably have averaged the left and right halves of the image already, so the two fits should be the same. Recommend using TRUE.
            USE_AU: this keyword is identical to UNIT in other methods. Can be 'pixel', 'au' or 'beam'.
            FLOOR_TO_0: whether or not to set everything below 0 to 0 when plotting. 
            H_OVER_R: if TRUE, plots y axis as h/r instead of h. '''
        
        # Smoothing
        if smooth:
            from scipy.signal import savgol_filter
            window_length = smooth
            polyorder = 3
            
            yup = handlelr(savgol_filter)(self.profile_up, window_length, polyorder)
            ymedian = handlelr(savgol_filter)(self.profile, window_length, polyorder)
            ydown = handlelr(savgol_filter)(self.profile_down, window_length, polyorder)
            
            yup2 = handlelr(savgol_filter)(self.profile_up2, window_length, polyorder)
            ymedian2 = handlelr(savgol_filter)(self.profile2, window_length, polyorder)
            ydown2 = handlelr(savgol_filter)(self.profile_down2, window_length, polyorder)
            
        else:
            yup = self.profile_up
            ymedian = self.profile
            ydown = self.profile_down
            
            yup2 = self.profile_up2
            ymedian2 = self.profile2
            ydown2 = self.profile_down2
        
        # Units
        if not use_au:
            r1 = self.rnew
            if self.image.is_fake_image:
                r2 = self.image.r_bounds_make
                heights_make = self.image.heights_make
            unit_label = 'pixels'
        elif use_au == True:
            r1 = self.rnew * self.image.scale
            if self.image.is_fake_image:
                r2 = self.image.r_bounds_make * self.image.scale
            if not h_over_r:
                yup = divlr(yup, 1/self.image.scale)
                ymedian = divlr(ymedian, 1/self.image.scale)
                ydown = divlr(ydown, 1/self.image.scale)
                yup2 = divlr(yup2, 1/self.image.scale)
                ymedian2 = divlr(ymedian2, 1/self.image.scale)
                ydown2 = divlr(ydown2, 1/self.image.scale)
                if self.image.is_fake_image:
                    heights_make = self.image.heights_make * self.image.scale
            unit_label = 'au'
        elif use_au == 'beam':
            assert self.image.beam_fwhm != None
            r1 = self.rnew / self.image.beam_fwhm
            if self.image.is_fake_image:
                r2 = self.image.r_bounds_make / self.image.beam_fwhm
            if not h_over_r:
                yup = divlr(yup, self.image.beam_fwhm)
                ymedian = divlr(ymedian, self.image.beam_fwhm)
                ydown = divlr(ydown, self.image.beam_fwhm)
                yup2 = divlr(yup2, self.image.beam_fwhm)
                ymedian2 = divlr(ymedian2, self.image.beam_fwhm)
                ydown2 = divlr(ydown2, self.image.beam_fwhm)
                if self.image.is_fake_image:
                    heights_make = self.image.heights_make / self.image.beam_fwhm
            unit_label = 'beam FWHMs'
        
        # Determine if plotting h or h/r
        if h_over_r:
            yup = divlr(yup, r1)
            ymedian = divlr(ymedian, r1)
            ydown = divlr(ydown, r1)
            if self.image.is_fake_image:
                r = (self.image.r_bounds_make[:-1] + self.image.r_bounds_make[1:])/2
                heights_make = self.image.heights_make / r
        
        # Don't plot regions with low flux
        for y in [yup, ymedian, ydown]:
            y['left'][self.image.radial.profile['left'] < F_lim * max(self.image.radial.profile['left'])] = np.nan
            y['right'][self.image.radial.profile['right'] < F_lim * max(self.image.radial.profile['right'])] = np.nan
        
        # floor to 0
        if 1:
            ymedian = floor(ymedian)
            yup = floor(yup)
            ydown = floor(ydown)
            
            ymedian2 = floor(ymedian2)
            yup2 = floor(yup2)
            ydown2 = floor(ydown2)
        
        # Plot fitted profiles
        plt.figure()
        if average:
            plt.plot(r1, meanlr(ymedian), color='C0', alpha=0.8, linewidth=2, label='Fitted from obs.')
            plt.fill_between(r1, meanlr(yup), meanlr(ydown), color='C0', alpha=0.5, linewidth=0)
            
            if self.extra_noise:
                plt.plot(r1, meanlr(ymedian2), color='C1', alpha=0.8, linewidth=2, label='Fitted from model')
                plt.fill_between(r1, meanlr(yup2), meanlr(ydown2), color='C1', alpha=0.3, linewidth=0)
        
        else:
            plt.plot(r1, ymedian['left'], alpha=0.8, linewidth=2, label='Left')
            plt.plot(r1, ymedian['right'], alpha=0.8, linewidth=2, label='Right')
            handlelr(plt.fill_between)(r1, yup, ydown, alpha=0.5)
        
        # Plot true profiles
        if self.image.is_fake_image:
            pltstep(r2, heights_make, color=colours[2], linestyle='--', label='True')
            plt.ylim([0, heights_make.max()*1.2])
        
        # Legend
        plt.legend(frameon=False)
        plt.title(f'{self.nrings} annuli, {self.image.noise_per_pixel/self.image.image.max()*100:.0f}% noise ({self.n_iterations} iterations)\n {self.inclination} inclination, {self.cut_height} cut height, {self.starting_height} starting height, {self.dh} dh')
        plt.xlabel(f'Radial distance ({unit_label})')
        if h_over_r:
            plt.ylabel('H/r')
        else:
            plt.ylabel(f'Height ({unit_label})')
        plt.tight_layout()
        plt.show()




from .lib import *
from .short_functions import *
from .ring_functions import *
from .linecut_functions import *


## Height calibrator and part generator
def make_generators(r_bounds, interp_heights, dim, RAPID_RINGS, kernel, cut_height=10, testing=True):
    '''This is run in preparation for FIT_HEIGHT2. 
    Returns
        HEIGHT_CALIBRATORS: a list containing the height-calibrating function of each annulus
        PART_GENERATORS: a list containing a function for each annulus that can quickly generate its midplane flux given a height. 
    
    Input
        R_BOUNDS: boundaries defining the annuli
        INTERP_HEIGHTS: the height values that are used to compute properly in order for the HEIGHT_CALIBRATORS and PART_GENERATORS to interpolate over. 
        DIM: dimensions of the image
        RAPID_RINGS: contains all the functions that let you rapidly generate images of rings. 
        KERNEL: kernel of the observation being fitted to. 
        CUT_HEIGHT: distance from midplaning defining the midplane region. 
        TESTING: always set this to TRUE. '''
    
    # Initialise
    nrings_make = len(r_bounds) - 1
    PART = handlelr(np.zeros)([len(interp_heights), nrings_make, dim//2])
    intp_range = handlelr(np.zeros)([nrings_make, 2])
    
    # Get data for each test height
    for i in range(len(interp_heights)):
        rings = RAPID_RINGS[interp_heights[i]](r_bounds, kernel)
        
        for j in range(nrings_make):
            ring = rings[j]
            part = make_linecut(ring, cut_height)
            PART['left'][i, j, :], PART['right'][i, j, :] = part['left'], part['right']

    # What height gives the right flux
    height_calibrators = createlr([], [])
    
    for j in range(nrings_make):
        maxparts = handlelr(np.zeros)(len(interp_heights))
        
        for i in range(len(interp_heights)):
            part = createlr(PART['left'][i, j, :], PART['left'][i, j, :])
            binned_part = bin_linecut(part, r_bounds)
            maxparts['left'][i], maxparts['right'][i] = binned_part['left'][j], binned_part['right'][j]
            
        f_height_left = interp1d(maxparts['left'], interp_heights, kind='linear')
        f_height_right = interp1d(maxparts['right'], interp_heights, kind='linear')
        height_calibrators['left'].append(f_height_left)
        height_calibrators['right'].append(f_height_right)
        
        intp_range['left'][j] = maxparts['left'][[0, -1]]
        intp_range['right'][j] = maxparts['right'][[0, -1]]

    # Generate linecut when height varied
    pixels, height = np.arange(dim//2), interp_heights
    part_generators = createlr([], [])
    
    for iring in range(nrings_make):
        part2d_left = np.array(PART['left'][:, iring, :])
        f_part2d_left = interp2d(pixels, height, part2d_left, kind='linear')
        part_generators['left'].append(f_part2d_left)
        
        part2d_right = np.array(PART['right'][:, iring, :])
        f_part2d_right = interp2d(pixels, height, part2d_right, kind='linear')
        part_generators['right'].append(f_part2d_right)
    
    if testing:
        return height_calibrators, part_generators, intp_range
    else:
        return height_calibrators, part_generators


## Apply ratio to flux, not height
def fit_height2(lim, radial_profile, rnew, r_bounds, height_calibrators, part_generators, H0, dim, intp_range=None):
    '''Fits the height given a set of annuli. 
    Returns
        H: a list containing the fitted height of each annulus
        LMOD: discretised midplane flux of the final fitted model
    
    Input
        LIM: discretised midplane flux of the image being fitted to 
        RADIAL_PROFILE: fitted surface brightness profile 
        RNEW: r-coordinates for RADIAL_PROFILE
        R_BOUNDS: boundaries defining the annuli
        HEIGHT_CALIBRATORS: see MAKE_GENERATORS
        PART_GENERATORS: see MAKE_GENERATORS
        H0: initial height at start of agorithm
        DIM: dimensions of the image
        INTP_RANGE: the "interpolation range", or the minimum and maximum fluxes of each annulus than can be mapped to its height. Mostly used for testing. '''
    
    if type(lim) is dict:
        left = fit_height2(lim['left'], radial_profile['left'], rnew, r_bounds, height_calibrators['left'], part_generators['left'], H0, dim, intp_range['left'])
        right = fit_height2(lim['right'], radial_profile['right'], rnew, r_bounds, height_calibrators['right'], part_generators['right'], H0, dim, intp_range['right'])
        H = createlr(left[0], right[0])
        lmod = createlr(left[1], right[1])
        return H, lmod
    
    # Interpolate from median radial profile
    delta = rnew[1] - rnew[0]
    radial_profile[radial_profile < 0] = 0
    F = bin_linecut(radial_profile * rnew, r_bounds/delta) / bin_linecut(rnew, r_bounds/delta)
    F2 = bin_linecut(radial_profile, r_bounds/delta)
    #print(F)
    #print(F2)
    
    # Parameters
    niterations = 100
    step_fraction = 0.3
    
    nrings = len(r_bounds) - 1
    H = np.zeros([niterations+1, nrings])
    lmod = np.zeros([niterations+1, nrings])
    pixels = np.arange(dim)
    
    # Initialise
    H[0, :] = H0
    previous_li = np.array([bin_linecut(part_generators[iring](pixels, H[0, iring]), r_bounds) for iring in range(nrings)])
    lmod[0, :] = weighted_sum(F, previous_li)
    
    
    for iter in range(1, niterations+1):
        # Before looping through rings, copy lmod from iter-1
        lmod[iter] = lmod[iter-1]
        
        for iring in range(nrings-1, -1, -1):
            # Calculate flux required of this ring if step_fraction of the residual flux comes from this ring
            target_flux = F[iring] * previous_li[iring, iring] + step_fraction * (lim[iring] - lmod[iter, iring])
            
            # Calculate height required to reach this flux
            try:
                if F[iring] > 0:
                    target_height = height_calibrators[iring](target_flux / F[iring])
                else:
                    #print(f'-> {iter}, {iring+1}, {H[iter-1, iring]} | Flux is {F[iring]}')
                    target_height = H[iter-1, iring]
            except ValueError:
                #print(f'-> {iter} r-{iring+1} | {H[iter-1, iring]:.1f}, {previous_li[iring, iring]:.1f} | {target_flux / F[iring]:.1f} [{intp_range[iring][0]:.1f}, {intp_range[iring][1]:.1f}]')
                if target_flux / F[iring] >= intp_range[iring][0]:
                    target_height = 0
                else:
                    target_height = H[iter-1, iring]
                '''
                if lmod[iter, iring] < lim[iring]:
                    target_height = H[iter-1, iring] - 1/step_fraction
                else:
                    target_height = H[iter-1, iring] + 1/step_fraction
                '''
            
            # Update height
            H[iter, iring] = target_height
            
            # Generate linecut for ring
            li = part_generators[iring](pixels, H[iter, iring])
            binned_li = bin_linecut(li, r_bounds)
            
            # Replace previous ring flux by current ring flux
            lmod[iter] = lmod[iter] - F[iring] * previous_li[iring] + F[iring] * binned_li
            previous_li[iring] = binned_li
    
    if 0:
        plt.figure()
        plt.plot(H)
        plt.show()
        
        plt.figure()
        plt.plot(lmod, alpha=0.8)
        plt.hlines(lim, 0, len(lmod)-1, ls='--')
        plt.show()
        
    return H, lmod


## Need to import these
import numpy as np
import matplotlib.pyplot as plt
from astropy.io import fits
import os
import time
from scipy.ndimage import gaussian_filter as blur
from scipy.ndimage import rotate
from scipy.interpolate import interp1d, interp2d
from scipy.ndimage import convolve

## Matplotlib GUI
import matplotlib
matplotlib.use('QT5Agg')
plt.rcParams['savefig.dpi'] = 300

## Aliases
pi = np.pi
colours = plt.rcParams['axes.prop_cycle'].by_key()['color']
read = fits.getdata

## Image processing functions
def rotate180(image):
    '''Rotates image by 180 deg about its center'''
    return np.flipud(np.fliplr(image))

def normalise(image):
    '''Sets brightest pixel in image to 1'''
    return image / np.max(image)

def reshape_image(image, new_ydim, new_xdim):
    '''Takes in an image, retuns an image with new dimensions.
    Pads image with 0's if new dimensions are bigger.
    Cuts out a region centred at the input image's centre if new dimensions are smaller.'''
    ydim, xdim = image.shape
    ycut, xcut = min(new_ydim, ydim), min(new_xdim, xdim)

    image_cut = image[ydim//2 - ycut//2 : ydim//2 + ycut//2, xdim//2 - xcut//2 : xdim//2 + xcut//2]

    new_image = np.zeros([new_ydim, new_xdim])
    new_image[new_ydim//2 - ycut//2 : new_ydim//2 + ycut//2, new_xdim//2 - xcut//2 : new_xdim//2 + xcut//2] = image_cut

    return new_image

def clip(image, lower, upper, rel=True):
    '''Sets everything in image bigger than UPPER to UPPER.
    Sets everything in image smaller than LOWER to LOWER.
    If REL is TRUE, LOWER and UPPER are interpreted as a fraction pf the brightest pixel.'''
    if rel:
        return np.clip(image, lower * np.max(image), upper * np.max(image))
    else:
        return np.clip(image, lower, upper)

## Plotting functions
def plot(image, x=None, title=None, cmap='viridis', n=True, flip=False, extent=None, marker='.-'):
    '''An umbrella plotting function that can plot 1D or 2D arrays.
    For example:
    >>> plot(image)
    >>> plot(y)
    >>> plot(x, y)'''
    newfigure = n
    if type(image) == str:
        if image[-5:] != '.fits':
            image += '.fits'
        image = fits.getdata(image)
    if newfigure:
        fig, ax = plt.subplots()
    image = np.array(image)
    if image.ndim == 2:
        if flip:
            image = np.fliplr(image)
        if extent is not None:
            plt.imshow(image, origin='lower', cmap=cmap, extent=extent)
        else:
            plt.imshow(image, origin='lower', cmap=cmap)
    elif image.ndim == 1:
        if x is not None:
            plt.plot(image, x, marker)
        else:
            plt.plot(image, marker)
    if title:
        plt.title(title)
    plt.show()
    if newfigure:
        return fig, ax

def cliplot(image, lo=None, hi=None):
    '''Plots the image after clipping the image between LO and HI.
    If only one bound is given, it is inperpreted as HI, and LO is set to 0.
    LO and HI are relative to the brightest pixel.'''
    if type(image) is str:
        if image[-5:] != '.fits':
            image += '.fits'
        image = fits.getdata(image)
    if hi is None:
        if lo is None:
            lo, hi = 0, 1
        else:
            lo, hi = 0, lo
    plot(clip(image, lo, hi))

def c():
    '''Close all MATPLOTLIB windows.'''
    plt.close('all')

